#!/usr/bin/python3
# -*- coding: utf-8 -*-
# key-mapper - GUI for device specific keyboard mappings
# Copyright (C) 2020 sezanzeb <proxima@hip70890b.de>
#
# This file is part of key-mapper.
#
# key-mapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# key-mapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with key-mapper.  If not, see <https://www.gnu.org/licenses/>.


import sys
import time
import grp
import os
import unittest
import evdev
from evdev.events import EV_KEY, EV_ABS
import json
from unittest.mock import patch
from importlib.util import spec_from_loader, module_from_spec
from importlib.machinery import SourceFileLoader

import gi
import shutil
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from keymapper.state import custom_mapping, system_mapping, XMODMAP_FILENAME
from keymapper.paths import CONFIG_PATH, get_preset_path
from keymapper.config import config, WHEEL, MOUSE
from keymapper.dev.reader import keycode_reader
from keymapper.gtk.row import to_string
from keymapper.dev import permissions

from tests.test import tmp, pending_events, InputEvent, \
    uinput_write_history_pipe, cleanup


def gtk_iteration():
    """Iterate while events are pending."""
    while Gtk.events_pending():
        Gtk.main_iteration()


# iterate a few times when Gtk.main() is called, but don't block
# there and just continue to the tests while the UI becomes
# unresponsive
Gtk.main = gtk_iteration

# doesn't do much except avoid some Gtk assertion error, whatever:
Gtk.main_quit = lambda: None


def launch(argv=None):
    """Start key-mapper-gtk with the command line argument array argv."""
    custom_mapping.empty()

    bin_path = os.path.join(os.getcwd(), 'bin', 'key-mapper-gtk')

    if not argv:
        argv = ['-d']

    with patch.object(sys, 'argv', [''] + [str(arg) for arg in argv]):
        loader = SourceFileLoader('__main__', bin_path)
        spec = spec_from_loader('__main__', loader)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

    gtk_iteration()

    return module.window


class FakeDropdown(Gtk.ComboBoxText):
    def __init__(self, name):
        self.name = name

    def get_active_text(self):
        return self.name

    def get_active_id(self):
        return self.name


class TestIntegration(unittest.TestCase):
    """For tests that use the window.

    Try to modify the configuration only by calling functions of the window.
    """
    def setUp(self):
        self.window = launch()

    def tearDown(self):
        self.window.on_apply_system_layout_clicked(None)
        gtk_iteration()
        self.window.on_close()
        self.window.window.destroy()
        gtk_iteration()
        cleanup()

    def get_rows(self):
        return self.window.get('key_list').get_children()

    def test_show_device_mapping_status(self):
        # this function may not return True, otherwise the timeout
        # runs forever
        self.assertFalse(self.window.show_device_mapping_status())

    def test_autoload(self):
        self.window.on_autoload_switch(None, False)
        self.assertFalse(config.is_autoloaded(
            self.window.selected_device,
            self.window.selected_preset
        ))

        self.window.on_select_device(FakeDropdown('device 1'))
        gtk_iteration()
        self.assertFalse(self.window.get('preset_autoload_switch').get_active())

        # select a preset for the first device
        self.window.get('preset_autoload_switch').set_active(True)
        gtk_iteration()
        self.assertTrue(self.window.get('preset_autoload_switch').get_active())
        self.assertTrue(config.is_autoloaded('device 1', 'new preset'))
        self.assertFalse(config.is_autoloaded('device 2', 'new preset'))
        self.assertListEqual(
            list(config.iterate_autoload_presets()),
            [('device 1', 'new preset')]
        )

        # switch the preset, the switch should be correct and the config
        # not changed.
        self.window.on_create_preset_clicked(None)
        gtk_iteration()
        self.assertEqual(self.window.selected_preset, 'new preset 2')
        self.assertFalse(self.window.get('preset_autoload_switch').get_active())
        self.assertTrue(config.is_autoloaded('device 1', 'new preset'))

        # select a preset for the second device
        self.window.on_select_device(FakeDropdown('device 2'))
        self.window.get('preset_autoload_switch').set_active(True)
        gtk_iteration()
        self.assertTrue(config.is_autoloaded('device 1', 'new preset'))
        self.assertTrue(config.is_autoloaded('device 2', 'new preset'))
        self.assertListEqual(
            list(config.iterate_autoload_presets()),
            [('device 1', 'new preset'), ('device 2', 'new preset')]
        )

        # disable autoloading for the second device
        self.window.get('preset_autoload_switch').set_active(False)
        gtk_iteration()
        self.assertTrue(config.is_autoloaded('device 1', 'new preset'))
        self.assertFalse(config.is_autoloaded('device 2', 'new preset'))
        self.assertListEqual(
            list(config.iterate_autoload_presets()),
            [('device 1', 'new preset')]
        )

    def test_select_device(self):
        # creates a new empty preset when no preset exists for the device
        self.window.on_select_device(FakeDropdown('device 1'))
        custom_mapping.change((EV_KEY, 50, 1), 'q')
        custom_mapping.change((EV_KEY, 51, 1), 'u')
        custom_mapping.change((EV_KEY, 52, 1), 'x')
        self.assertEqual(len(custom_mapping), 3)
        self.window.on_select_device(FakeDropdown('device 2'))
        self.assertEqual(len(custom_mapping), 0)
        # it creates the file for that right away. It may have been possible
        # to write it such that it doesn't (its empty anyway), but it does,
        # so use that to test it in more detail.
        path = get_preset_path('device 2', 'new preset')
        self.assertTrue(os.path.exists(path))
        with open(path, 'r') as file:
            preset = json.load(file)
            self.assertEqual(len(preset['mapping']), 0)

    def test_can_start(self):
        self.assertIsNotNone(self.window)
        self.assertTrue(self.window.window.get_visible())

    def test_row_keycode_to_string(self):
        # not an integration test, but I have all the row tests here already
        self.assertEqual(to_string(EV_KEY, evdev.ecodes.KEY_9, 1), '9')
        self.assertEqual(to_string(EV_KEY, evdev.ecodes.KEY_SEMICOLON, 1), 'SEMICOLON')
        self.assertEqual(to_string(EV_ABS, evdev.ecodes.ABS_HAT0X, -1), 'ABS_HAT0X L')
        self.assertEqual(to_string(EV_ABS, evdev.ecodes.ABS_HAT0X, 1), 'ABS_HAT0X R')
        self.assertEqual(to_string(EV_KEY, evdev.ecodes.BTN_A, 1), 'BTN_A')

    def test_row_simple(self):
        rows = self.window.get('key_list').get_children()
        self.assertEqual(len(rows), 1)

        row = rows[0]

        row.set_new_keycode(None)
        self.assertIsNone(row.get_keycode())
        self.assertEqual(len(custom_mapping), 0)
        self.assertEqual(row.keycode_input.get_label(), None)

        row.set_new_keycode((EV_KEY, 30, 1))
        self.assertEqual(len(custom_mapping), 0)
        self.assertEqual(row.get_keycode(), (EV_KEY, 30, 1))
        # this is KEY_A in linux/input-event-codes.h,
        # but KEY_ is removed from the text
        self.assertEqual(row.keycode_input.get_label(), 'A')

        row.set_new_keycode((EV_KEY, 30, 1))
        self.assertEqual(len(custom_mapping), 0)
        self.assertEqual(row.get_keycode(), (EV_KEY, 30, 1))

        time.sleep(0.1)
        gtk_iteration()
        self.assertEqual(len(self.window.get('key_list').get_children()), 1)

        row.character_input.set_text('Shift_L')
        self.assertEqual(len(custom_mapping), 1)

        time.sleep(0.1)
        gtk_iteration()
        self.assertEqual(len(self.window.get('key_list').get_children()), 2)

        self.assertEqual(custom_mapping.get_character((EV_KEY, 30, 1)), 'Shift_L')
        self.assertEqual(row.get_character(), 'Shift_L')
        self.assertEqual(row.get_keycode(), (EV_KEY, 30, 1))

    def change_empty_row(self, key, char, code_first=True, expect_success=True):
        """Modify the one empty row that always exists.

        Utility function for other tests.

        Parameters
        ----------
        key : int, int, int
            type, code, value
        code_first : boolean
            If True, the code is entered and then the character.
            If False, the character is entered first.
        expect_success : boolean
            If this change on the empty row is going to result in a change
            in the mapping eventually. False if this change is going to
            cause a duplicate.
        """
        # wait for the window to create a new empty row if needed
        time.sleep(0.1)
        gtk_iteration()

        # find the empty row
        rows = self.get_rows()
        row = rows[-1]
        self.assertIsNone(row.keycode_input.get_label())
        self.assertEqual(row.character_input.get_text(), '')
        self.assertNotIn('changed', row.get_style_context().list_classes())

        if char and not code_first:
            # set the character to make the new row complete
            self.assertIsNone(row.get_character())
            row.character_input.set_text(char)
            self.assertEqual(row.get_character(), char)

        self.window.window.set_focus(row.keycode_input)

        if key:
            # modifies the keycode in the row not by writing into the input,
            # but by sending an event
            keycode_reader._pipe[1].send(InputEvent(*key))
            time.sleep(0.1)
            gtk_iteration()
            if expect_success:
                self.assertEqual(row.get_keycode(), key)
                css_classes = row.get_style_context().list_classes()
                self.assertIn('changed', css_classes)

        if not expect_success:
            self.assertIsNone(row.get_keycode())
            self.assertIsNone(row.get_character())
            self.assertNotIn('changed', row.get_style_context().list_classes())
            return row

        if char and code_first:
            # set the character to make the new row complete
            self.assertIsNone(row.get_character())
            row.character_input.set_text(char)
            self.assertEqual(row.get_character(), char)

        return row

    def test_rows(self):
        """Comprehensive test for rows."""
        # how many rows there should be in the end
        num_rows_target = 3

        ev_1 = (EV_KEY, 10, 1)
        ev_2 = (EV_ABS, evdev.ecodes.ABS_HAT0X, -1)

        """edit"""

        # add two rows by modifiying the one empty row that exists
        self.change_empty_row(ev_1, 'a', code_first=False)
        self.change_empty_row(ev_2, 'k(b).k(c)')

        # one empty row added automatically again
        time.sleep(0.1)
        gtk_iteration()
        self.assertEqual(len(self.get_rows()), num_rows_target)

        self.assertEqual(custom_mapping.get_character(ev_1), 'a')
        self.assertEqual(custom_mapping.get_character(ev_2), 'k(b).k(c)')
        self.assertTrue(custom_mapping.changed)

        """save"""

        self.window.on_save_preset_clicked(None)
        for row in self.get_rows():
            css_classes = row.get_style_context().list_classes()
            self.assertNotIn('changed', css_classes)

        self.assertFalse(custom_mapping.changed)

        """edit first row"""

        # now change the first row and it should turn blue,
        # but the other should remain unhighlighted
        row = self.get_rows()[0]
        row.character_input.set_text('c')
        self.assertIn('changed', row.get_style_context().list_classes())
        for row in self.get_rows()[1:]:
            css_classes = row.get_style_context().list_classes()
            self.assertNotIn('changed', css_classes)

        self.assertEqual(custom_mapping.get_character(ev_1), 'c')
        self.assertEqual(custom_mapping.get_character(ev_2), 'k(b).k(c)')
        self.assertTrue(custom_mapping.changed)

        """add duplicate"""

        # try to add a duplicate keycode, it should be ignored
        self.change_empty_row(ev_2, 'd', expect_success=False)
        self.assertEqual(custom_mapping.get_character(ev_2), 'k(b).k(c)')
        # and the number of rows shouldn't change
        self.assertEqual(len(self.get_rows()), num_rows_target)

    def test_hat0x(self):
        # it should be possible to add all of them
        ev_1 = (EV_ABS, evdev.ecodes.ABS_HAT0X, -1)
        ev_2 = (EV_ABS, evdev.ecodes.ABS_HAT0X, 1)
        ev_3 = (EV_ABS, evdev.ecodes.ABS_HAT0Y, -1)
        ev_4 = (EV_ABS, evdev.ecodes.ABS_HAT0Y, 1)

        self.change_empty_row(ev_1, 'a')
        self.change_empty_row(ev_2, 'b')
        self.change_empty_row(ev_3, 'c')
        self.change_empty_row(ev_4, 'd')

        self.assertEqual(custom_mapping.get_character(ev_1), 'a')
        self.assertEqual(custom_mapping.get_character(ev_2), 'b')
        self.assertEqual(custom_mapping.get_character(ev_3), 'c')
        self.assertEqual(custom_mapping.get_character(ev_4), 'd')
        self.assertTrue(custom_mapping.changed)

        # and trying to add them as duplicate rows will be ignored for each
        # of them
        self.change_empty_row(ev_1, 'e', expect_success=False)
        self.change_empty_row(ev_2, 'f', expect_success=False)
        self.change_empty_row(ev_3, 'g', expect_success=False)
        self.change_empty_row(ev_4, 'h', expect_success=False)

        self.assertEqual(custom_mapping.get_character(ev_1), 'a')
        self.assertEqual(custom_mapping.get_character(ev_2), 'b')
        self.assertEqual(custom_mapping.get_character(ev_3), 'c')
        self.assertEqual(custom_mapping.get_character(ev_4), 'd')
        self.assertTrue(custom_mapping.changed)

    def test_remove_row(self):
        """Comprehensive test for rows 2."""
        # sleeps are added to be able to visually follow and debug the test
        # add two rows by modifiying the one empty row that exists
        row_1 = self.change_empty_row((EV_KEY, 10, 1), 'a')
        row_2 = self.change_empty_row((EV_KEY, 11, 1), 'b')
        row_3 = self.change_empty_row(None, 'c')

        # no empty row added because one is unfinished
        time.sleep(0.2)
        gtk_iteration()
        self.assertEqual(len(self.get_rows()), 3)

        self.assertEqual(custom_mapping.get_character((EV_KEY, 11, 1)), 'b')

        def remove(row, code, char, num_rows_after):
            if code is not None and char is not None:
                self.assertEqual(custom_mapping.get_character((EV_KEY, code, 1)), char)

            self.assertEqual(row.get_character(), char)
            if code is None:
                self.assertIsNone(row.get_keycode())
            else:
                self.assertEqual(row.get_keycode(), (EV_KEY, code, 1))
            row.on_delete_button_clicked()
            time.sleep(0.2)
            gtk_iteration()
            self.assertIsNone(row.get_keycode())
            self.assertIsNone(row.get_character())
            self.assertIsNone(custom_mapping.get_character((EV_KEY, code, 1)))
            self.assertEqual(len(self.get_rows()), num_rows_after)

        remove(row_1, 10, 'a', 2)
        remove(row_2, 11, 'b', 1)
        # there is no empty row at the moment, so after removing that one,
        # which is the only row, one empty row will be there. So the number
        # of rows won't change.
        remove(row_3, None, 'c', 1)

    def test_rename_and_save(self):
        custom_mapping.change((EV_KEY, 14, 1), 'a', None)
        self.assertEqual(self.window.selected_preset, 'new preset')
        self.window.on_save_preset_clicked(None)
        self.assertEqual(custom_mapping.get_character((EV_KEY, 14, 1)), 'a')

        custom_mapping.change((EV_KEY, 14, 1), 'b', None)
        self.window.get('preset_name_input').set_text('asdf')
        self.window.on_save_preset_clicked(None)
        self.assertEqual(self.window.selected_preset, 'asdf')
        self.assertTrue(os.path.exists(f'{CONFIG_PATH}/presets/device 1/asdf.json'))
        self.assertEqual(custom_mapping.get_character((EV_KEY, 14, 1)), 'b')

    def test_check_macro_syntax(self):
        status = self.window.get('status_bar')

        custom_mapping.change((EV_KEY, 9, 1), 'k(1))', None)
        self.window.on_save_preset_clicked(None)
        tooltip = status.get_tooltip_text().lower()
        self.assertIn('brackets', tooltip)

        custom_mapping.change((EV_KEY, 9, 1), 'k(1)', None)
        self.window.on_save_preset_clicked(None)
        tooltip = status.get_tooltip_text().lower()
        self.assertNotIn('brackets', tooltip)
        self.assertIn('saved', tooltip)

        self.assertEqual(custom_mapping.get_character((EV_KEY, 9, 1)), 'k(1)')

    def test_select_device_and_preset(self):
        # created on start because the first device is selected and some empty
        # preset prepared.
        self.assertTrue(os.path.exists(f'{CONFIG_PATH}/presets/device 1/new preset.json'))
        self.assertEqual(self.window.selected_device, 'device 1')
        self.assertEqual(self.window.selected_preset, 'new preset')

        # create another one
        self.window.on_create_preset_clicked(None)
        gtk_iteration()
        self.assertTrue(os.path.exists(f'{CONFIG_PATH}/presets/device 1/new preset.json'))
        self.assertTrue(os.path.exists(f'{CONFIG_PATH}/presets/device 1/new preset 2.json'))
        self.assertEqual(self.window.selected_preset, 'new preset 2')

        self.window.on_select_preset(FakeDropdown('new preset'))
        gtk_iteration()
        self.assertEqual(self.window.selected_preset, 'new preset')

        self.assertListEqual(
            sorted(os.listdir(f'{CONFIG_PATH}/presets/device 1')),
            sorted(['new preset.json', 'new preset 2.json'])
        )

        # now try to change the name
        self.window.get('preset_name_input').set_text('abc 123')
        gtk_iteration()
        self.assertEqual(self.window.selected_preset, 'new preset')
        self.assertFalse(os.path.exists(f'{CONFIG_PATH}/presets/device 1/abc 123.json'))
        custom_mapping.change((EV_KEY, 10, 1), '1', None)
        self.window.on_save_preset_clicked(None)
        gtk_iteration()
        self.assertEqual(self.window.selected_preset, 'abc 123')
        self.assertTrue(os.path.exists(f'{CONFIG_PATH}/presets/device 1/abc 123.json'))
        self.assertListEqual(
            sorted(os.listdir(os.path.join(CONFIG_PATH, 'presets'))),
            sorted(['device 1'])
        )
        self.assertListEqual(
            sorted(os.listdir(f'{CONFIG_PATH}/presets/device 1')),
            sorted(['abc 123.json', 'new preset 2.json'])
        )

    def test_gamepad_config(self):
        # select a device that is not a gamepad
        self.window.on_select_device(FakeDropdown('device 1'))
        self.assertFalse(self.window.get('gamepad_config').is_visible())

        # select a gamepad
        self.window.on_select_device(FakeDropdown('gamepad'))
        self.assertTrue(self.window.get('gamepad_config').is_visible())

        # set stuff
        self.window.get('left_joystick_purpose').set_active_id(WHEEL)
        self.window.get('right_joystick_purpose').set_active_id(WHEEL)
        joystick_mouse_speed = 5
        self.window.get('joystick_mouse_speed').set_value(joystick_mouse_speed)

        # it should be stored in custom_mapping, which overwrites the
        # global config
        config.set('gamepad.joystick.left_purpose', MOUSE)
        config.set('gamepad.joystick.right_purpose', MOUSE)
        config.set('gamepad.joystick.pointer_speed', 50)
        left_purpose = custom_mapping.get('gamepad.joystick.left_purpose')
        right_purpose = custom_mapping.get('gamepad.joystick.right_purpose')
        pointer_speed = custom_mapping.get('gamepad.joystick.pointer_speed')
        self.assertEqual(left_purpose, WHEEL)
        self.assertEqual(right_purpose, WHEEL)
        self.assertEqual(pointer_speed, 2 ** joystick_mouse_speed)

        # select a device that is not a gamepad again
        self.window.on_select_device(FakeDropdown('device 1'))
        self.assertFalse(self.window.get('gamepad_config').is_visible())

    def test_start_injecting(self):
        keycode_from = 9
        keycode_to = 200

        self.change_empty_row((EV_KEY, keycode_from, 1), 'a')
        system_mapping.clear()
        system_mapping._set('a', keycode_to)

        pending_events['device 2'] = [
            InputEvent(evdev.events.EV_KEY, keycode_from, 1),
            InputEvent(evdev.events.EV_KEY, keycode_from, 0)
        ]

        custom_mapping.save(get_preset_path('device 2', 'foo preset'))

        # use only the manipulated system_mapping
        os.remove(os.path.join(tmp, XMODMAP_FILENAME))

        self.window.selected_device = 'device 2'
        self.window.selected_preset = 'foo preset'
        self.window.on_apply_preset_clicked(None)

        # the integration tests will cause the injection to be started as
        # processes, as intended. Luckily, recv will block until the events
        # are handled and pushed.

        # Note, that pushing events to pending_events won't work anymore
        # from here on because the injector processes memory cannot be
        # modified from here.

        event = uinput_write_history_pipe[0].recv()
        self.assertEqual(event.type, evdev.events.EV_KEY)
        self.assertEqual(event.code, keycode_to)
        self.assertEqual(event.value, 1)

        event = uinput_write_history_pipe[0].recv()
        self.assertEqual(event.type, evdev.events.EV_KEY)
        self.assertEqual(event.code, keycode_to)
        self.assertEqual(event.value, 0)

    def test_stop_injecting(self):
        keycode_from = 16
        keycode_to = 90

        self.change_empty_row((EV_KEY, keycode_from, 1), 't')
        system_mapping.clear()
        system_mapping._set('t', keycode_to)

        # not all of those events should be processed, since that takes some
        # time due to time.sleep in the fakes and the injection is stopped.
        pending_events['device 2'] = [InputEvent(1, keycode_from, 1)] * 100

        custom_mapping.save(get_preset_path('device 2', 'foo preset'))

        self.window.selected_device = 'device 2'
        self.window.selected_preset = 'foo preset'
        self.window.on_apply_preset_clicked(None)

        pipe = uinput_write_history_pipe[0]
        # block until the first event is available, indicating that
        # the injector is ready
        write_history = [pipe.recv()]

        # stop
        self.window.on_apply_system_layout_clicked(None)

        # try to receive a few of the events
        time.sleep(0.2)
        while pipe.poll():
            write_history.append(pipe.recv())

        len_before = len(write_history)
        self.assertLess(len(write_history), 50)

        # since the injector should not be running anymore, no more events
        # should be received after waiting even more time
        time.sleep(0.2)
        while pipe.poll():
            write_history.append(pipe.recv())
        self.assertEqual(len(write_history), len_before)


original_access = os.access
original_getgrnam = grp.getgrnam
original_can_read_devices = permissions.can_read_devices


class TestPermissions(unittest.TestCase):
    def tearDown(self):
        os.access = original_access
        os.getgrnam = original_getgrnam
        permissions.can_read_devices = original_can_read_devices

        if self.window is not None:
            self.window.on_close()
            self.window.window.destroy()
            gtk_iteration()
            self.window = None

        shutil.rmtree('/tmp/key-mapper-test')

    def test_fails(self):
        def fake():
            return ['error1', 'error2', 'error3']

        permissions.can_read_devices = fake

        self.window = launch()
        status = self.window.get('status_bar')

        tooltip = status.get_tooltip_text()
        self.assertIn('sudo', tooltip)
        self.assertIn('pkexec', tooltip)
        self.assertIn('error1', tooltip)
        self.assertIn('error2', tooltip)
        self.assertIn('error3', tooltip)

    def test_good(self):
        def fake():
            return []

        permissions.can_read_devices = fake

        self.window = launch()
        status = self.window.get('status_bar')

        self.assertIsNone(status.get_tooltip_text())


if __name__ == "__main__":
    unittest.main()
